"""Module to aid with data validation"""


def assert_positive_number(n, name):
    """Assert that an input is a positive number."""
    if type(n) not in [int, float]:
        raise ValueError(
            f"The value for '{name}' should be an integer or a float, not a {type(n)}.")
    if n < 0:
        raise ValueError(
            f"The value for '{name}' should be >= 0, not {n}")

def assert_strictly_positive_number(n, name):
    """Assert that an input is a strictly positive number."""
    if type(n) not in [int, float]:
        raise ValueError(
            f"The value for '{name}' should be an integer or a float, not a {type(n)}.")
    if n < 0:
        raise ValueError(
            f"The value for '{name}' should be > 0, not {n}")


def assert_number(n, name):
    """Assert that an input is a number."""
    if type(n) not in [int, float]:
        raise ValueError(
            f"The value for '{name}' should be an integer or a float, not a {type(n)}.")



"""Module to help draw plotly shapes, drawing and annotations"""

from plotly.subplots import make_subplots
import plotly.graph_objects as go
import numpy as np
from sympy import lambdify, sympify, sin, cos, oo
from sympy.abc import x
from math import radians


def draw_line(fig, angle, x_sup, length=-20, xoffset=0, yoffset=0,
              color='red', line_width=2, row=None, col=None):
    """Draw an anchored line on a plotly figure.

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append line shape to.
    angle : int
        Angle of the line from the x-axis. Angle uses standard mathematical
        cartesian convention.
    x_sup : int
        The x position for the line to be anchored to.
    length : int, optional
        the line length, by default -20
    xoffset : int, optional
        The x-offset of the start of the line from the anchor, by default 0
    yoffset : int, optional
        The y-offset of the start of the line from the anchor, by default 0
    color : str, optional
        Line color, by default 'red'
    line_width : int, optional
        Line width, by default 2
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.

    Returns
    -------
    plotly figure
        Returns the plotly figure passed into function with the arrowhead
        appended to it.
    """
    # Establish line start and end coordinates.
    x0 = xoffset
    y0 = yoffset
    x1 = x0 + int(sympify(length * cos(radians(angle))).evalf(2))
    y1 = y0 + int(sympify(length * sin(radians(angle))).evalf(2))

    # Create dictionary for shape object representing line.
    shape = dict(
        type="line",
        xref="x", yref="y",
        x0=x0, y0=y0, x1=x1, y1=y1,
        line_color=color, line_width=line_width,
        xsizemode='pixel', ysizemode='pixel',
        xanchor=x_sup, yanchor=0)

    # Append shape to plot or subplot
    if row and col:
        fig.add_shape(shape, row=row, col=col)
    else:
        fig.add_shape(shape)

    return fig


def draw_arrowhead(fig, angle, x_sup, length=5, xoffset=0, yoffset=0,
                   color='red', line_width=2, row=None, col=None):
    """Draw an anchored arrowhead on a plotly figure.

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append arrowhead shape to.
    angle : int
        Angle of the arrowhead from the x-axis. Angle uses standard
        mathematical cartesian convention.
    x_sup : int
        The x position for the arrowhead to be anchored to.
    length : int, optional
        the arrowhead length, by default 5
    xoffset : int, optional
        The x-offset of the start of the arrowhead from the anchor, by
        default 0
    yoffset : int, optional
        The y-offset of the start of the arrowhead from the anchor, by
        default 0
    color : str, optional
        arrowhead color, by default 'red'
    line_width : int, optional
        Line width, by default 2
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.

    Returns
    -------
    plotly figure
        Returns the plotly figure passed into function with the arrowhead
        appended to it.
    """
    # Holds lines 90 degrees apart to represent arrowhead. Constructed so 0
    # degrees is pointing right, follows conventions in documentation for angle

    # Angle conversion to allow for compatability with draw_line function
    a1 = 225 + angle
    a2 = 135 + angle

    # Append line to figure (half of arrowhead)
    fig = draw_line(
        fig,
        angle=a1,
        x_sup=x_sup,
        length=length,
        xoffset=xoffset,
        yoffset=yoffset,
        color=color,
        line_width=line_width,
        row=row,
        col=col)

    # Append line to figure (half of arrowhead)
    fig = draw_line(
        fig,
        angle=a2,
        x_sup=x_sup,
        length=length,
        xoffset=xoffset,
        yoffset=yoffset,
        color=color,
        line_width=line_width,
        row=row,
        col=col)

    return fig


def draw_arrow(fig, angle, force, x_sup, xoffset=0, yoffset=0, color='red',
               line_width=2, arrowhead=5, arrowlength=30, show_values=True,row=None, col=None):
    """Draw an anchored arrow on a plotly figure.

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append arrow shape to.
    angle : int
        Angle of the arrow from the x-axis. Angle uses standard
        mathematical cartesian convention.
    force: int
        force that the arrow will represent. Only need to know whether it
        is positive or negative, but it generally is easiest to just parse
        the whole force in.
    x_sup : int
        The x position for the arrow to be anchored to.
    xoffset : int, optional
        The x-offset of the start of the arrow from the anchor, by default 0
    yoffset : int, optional
        The y-offset of the start of the arrow from the anchor, by default 0
    color : str, optional
        arrow color, by default 'red'
    line_width : int, optional
        Line width, by default 2
    arrowhead : int, optional
        Size of the arrowhead lines, by default 5
    arrowlength: int, optional
        length of the arrow line, by default 30
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.



    Returns
    -------
    plotly figure
        Returns the plotly figure passed into function with the arrow
        appended to it.
    """

    # Factor to switch arrow direction based on force sign
    if force > 0:
        d = 1
    else:
        d = -1

    # Draw arrowhead for force
    fig = draw_arrowhead(
        fig,
        angle,
        x_sup,
        length=arrowhead * d,
        xoffset=xoffset,
        yoffset=yoffset,
        color=color,
        line_width=line_width,
        row=row,
        col=col)

    # Draw arrowline for force
    fig = draw_line(
        fig,
        angle,
        x_sup,
        length=-1 * arrowlength * d,
        xoffset=xoffset,
        yoffset=yoffset,
        color=color,
        line_width=line_width,
        row=row,
        col=col)
    if show_values:
        # determine start and end of arrow
        x0 = xoffset + x_sup
        y0 = yoffset
        x1 = (x0 + int(sympify(-arrowlength* d * cos(radians(angle))).evalf(2)))*1.1
        y1 = (y0 + int(sympify(-arrowlength * d * sin(radians(angle))).evalf(2)))*1.3
        
        # make so text doesnt intersect x axis
        if abs(y1)<5:
            if y1 >= 0:
                y1=10
            else:
                y1=-10

        annotation = dict(
            xref="x", yref="y",
            x=x0,
            y=y0,
            xshift=x1,
            yshift=y1,
            text=force,
            font_color=color,
            showarrow=False,
            )

        # Append shape to plot or subplot
        if row and col:
            fig.add_annotation(annotation, row=row, col=col)
        else:
            fig.add_annotation(annotation)

    return fig


def draw_support_triangle(fig, x_sup, orientation="up", row=None, col=None):
    """Draw an anchored triangle on a plotly figure.

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append arrow shape to.
    x_sup : int
        The x position for the arrow to be anchored to.
    orientation : 'up' or 'right, optional
        direction that the arrow faces, by default "up"
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.


    Returns
    -------
    plotly figure
        Returns the plotly figure passed into function with the triangle
        appended to it.
    """

    if orientation in ['up', 'right']:

        # Define the triangle as a point using the scatter marker.
        trace = go.Scatter(
            x=[x_sup],
            y=[0],
            fill="toself",
            showlegend=False,
            mode="markers",
            name='Support',
            marker=dict(
                symbol="arrow-" + orientation,
                size=10,
                color='blue'),
            hovertemplate=None,
            hoverinfo="skip")

        # Append trace to plot or subplot
        if row and col:
            fig.add_trace(trace, row=row, col=col)
        else:
            fig.add_trace(trace)

    return fig


def draw_support_rectangle(fig, x_sup, orientation="up", row=None, col=None):
    """Draw an anchored rectangle on a plotly figure.

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append arrow shape to.
    x_sup : int
        The x position for the arrow to be anchored to.
    orientation : 'up' or 'right, optional
        direction that the arrow faces, by default "up"
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.


    Returns
    -------
    plotly figure
        Returns the plotly figure passed into function with the rectangle
        appended to it.
    """
    if orientation in ['up', 'right']:

        # rectangle is offset by -2 so that when a thickness of 4 is used
        # points are defined for a line length of 10
        if orientation == 'up':
            x0, x1, y0, y1 = -5, 5, -2, -2
        elif orientation == 'right':
            x0, x1, y0, y1 = -2, -2, -5, 5

        # Create dictionary for shape object representing rectangle.
        shape = dict(
            type="rect",
            xref="x", yref="y",
            x0=x0, y0=y0, x1=x1, y1=y1,
            line_color="blue",
            line_width=4,
            xsizemode='pixel',
            ysizemode='pixel',
            fillcolor='blue',
            xanchor=x_sup,
            yanchor=0)

        # Append shape to plot or subplot
        if row and col:
            fig.add_shape(shape, row=row, col=col)
        else:
            fig.add_shape(shape)

    return fig


def draw_moment(fig, moment, x_sup, color='magenta', show_values=True, row=None,
                col=None):
    """Draw a moment (torque) shape (circular arrow) on a plotly figure.

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append circular arrow shape to.
    x_sup : int
        The x position for the circular arrow to be anchored to.
    direction : 'clockwise' or 'anti-clockwise', optional
        direction for circular arrow to point in, default 'clockwise
    color : str, optional
        color of circular arrow, default 'red'
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.

    Returns
    -------
    plotly figure
        Returns the plotly figure passed into function with the circular
        arrow appended to it.
    """
    # Choose symbol based on direction given
    if moment < 0:
        d = "⭮"
    elif moment > 0:
        d = "⭯"
    else:
        return fig

    # Create dictionary for annotation object representing moment.
    annotation = dict(
        x=x_sup, y=0,
        text=d,
        showarrow=False,
        yshift=0,
        font_size=26,
        font_color=color,
    )

    # Append shape to plot or subplot
    if row and col:
        fig.add_annotation(annotation, row=row, col=col)
    else:
        fig.add_annotation(annotation)

    # Add text if show_values is true
    if show_values:

        annotation = dict(
            xref="x", yref="y",
            x=x_sup,
            y=0,
            yshift=-20,
            text=moment,
            font_color=color,
            showarrow=False,
            )

        # Append shape to plot or subplot
        if row and col:
            fig.add_annotation(annotation, row=row, col=col)
        else:
            fig.add_annotation(annotation)

    return fig


def draw_force(fig, load, row=None, col=None):
    """Draw a force (for load or reaction) on a plotly figure

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append force representation to.
    load : instance of a load class
        force to be represented on figure
    color : str, optional
        color of force representation, default 'red'
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.

    Returns
    -------
    plotly figure
        Returns the plotly figure passed into function with the force
        representation appended to it.
    """
    # Get namedtuple name to check the type of load.  (Is used as an
    # alternative to isinstance method in order to avoid importing the
    # load classes to avoid circular dependency.)
    load_type = str(load).split('(')[0]

    if load_type == 'PointTorque':
        moment, x_sup = load
        fig = draw_moment(
            fig,
            moment,
            x_sup,
            row=row,
            col=col)

    elif load_type == 'PointLoadH':
        force, x_sup = load
        angle = 0

        fig = draw_arrow(
            fig,
            angle,
            force,
            x_sup,
            row=row,
            col=col)

    elif load_type == 'PointLoadV':
        force, x_sup = load
        angle = 90

        fig = draw_arrow(
            fig,
            angle,
            force,
            x_sup,
            row=row,
            col=col)

    elif load_type == 'PointLoad':
        force, x_sup, angle = load

        fig = draw_arrow(
            fig,
            angle,
            force,
            x_sup,
            row=row,
            col=col)

    elif load_type == 'DistributedLoadV':
        # need to know sign for each side.
        # draw each function normalised to 1. ie the max is always 1.
        # evaluate force at left and force at right to plot
        expr, interval = load
        x0, x1 = interval
        expr = sympify(expr)
        # numpy array for x positions closely spaced (allow for graphing)
        x_vec = np.linspace(x0, x1, int(min((x1 - x0) * 100 + 1, 1e4)))
        y_lam = lambdify(x, expr, "numpy")
        y_vec = np.array([y_lam(t) for t in x_vec])

        largest = abs(max(y_vec, key=abs))
        # normalise  , use -1 to flip direction so matches arrow direction
        y_vec = (-1) * y_vec / largest

        # Create trace object for graph of distributed force
        trace = go.Scatter(
            x=x_vec.tolist(),
            y=y_vec.tolist(),
            mode='lines',
            line=dict(
                color='darkgreen',
                width=1),
            fill='tozeroy',
            name='DistributedLoadV',
            hovertemplate="",
            hoverinfo="skip")

        # Append to plot or subplot
        if row and col:
            fig.add_trace(trace, row=row, col=col)
        else:
            fig.add_trace(trace)

        # draw arrow for left force and right force (if larger than 2% of
        # max load)
        angle = 90
        if abs(y_vec[0]) > 0.02:
            fig = draw_arrow(
                fig,
                angle,
                -y_vec[0] * largest,
                x_vec[0],
                color='green',
                arrowlength=30*abs(y_vec[0]),
                row=row, 
                col=col)

        if abs(y_vec[-1]) > 0.02:
            fig = draw_arrow(
                fig,
                angle,
                -y_vec[-1]* largest,
                x_vec[-1],
                color='green',
                arrowlength=30*abs(y_vec[-1]),
                row=row,
                col=col)

    return fig


def draw_load_hoverlabel(fig, load, row=None, col=None):
    """Draw a load hoverlabel on a plotly figure

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append force representation to.
    load : instance of a load class
        force to be represented on figure
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.


    Returns
    -------
    plotly figure
        plotly figure to append hoverlabel representation to.
    """

    # Get namedtuple name to check the type of load.  (Is used as an
    # alternative to isinstance method in order to avoid importing the
    # load classes to avoid circular dependency.)
    load_type = str(load).split('(')[0]

    y_sup = 0

    if load_type in ['PointLoad', 'PointLoadH', 'PointLoadV', 'PointTorque']:
        x_sup = load[1]
        color = 'red'
        meta = [load[0], load[1]]  # load, position

        if load_type == 'PointTorque':
            color = 'magenta'
            hovertemplate = 'x: %{meta[1]} m<br>Moment: %{meta[0]} kN.m'
            name = 'Point<br>Torque'
        elif load_type == 'PointLoad':
            meta.append(load[2])
            hovertemplate = 'x: %{meta[1]} m<br>Force: %{meta[0]} \
                kN<br>Angle: %{meta[2]} deg'
            name = 'Point<br>Load'
        elif load_type == 'PointLoadV':
            hovertemplate = 'x: %{meta[1]} m<br>Force: %{meta[0]} \
                kN<br>Direction: Vertical'
            name = 'Point<br>LoadV'
        else:
            hovertemplate = 'x: %{meta[1]} m<br>Force: %{meta[0]} \
                kN<br>Angle: Horizontal'
            name = 'Point<br>LoadH'

        # Define hoverlabel as a marker with 0 opacity and a hovertemplate that
        # relies on meta data field
        trace = go.Scatter(
            x=[x_sup], y=[y_sup],
            showlegend=False, mode="markers",
            name=name,
            meta=meta,
            marker=dict(symbol="triangle-up", size=10, color=color),
            hovertemplate=hovertemplate,
            hoverinfo="skip",
            opacity=0
        )

        # Append hoverlabel to plot or subplot
        if row and col:
            fig.add_trace(trace, row=row, col=col)
        else:
            fig.add_trace(trace)

    # Else is distributed load type, hoverlabel needed for arrow at each side
    # of function
    else:
        expr, interval = load
        x0, x1 = interval
        expr = sympify(expr)
        # numpy array for x positions closely spaced (allow for graphing)
        
        y_lam = lambdify(x, expr, "numpy")

        name = 'Distributed<br>Load'
        color = 'green'
        y_sup = 1

        meta = [x0, x1, y_lam(x0), y_lam(x1)]

        trace_1 = go.Scatter(
            x=[x0], y=[0],
            showlegend=False, mode="markers",
            name=name,
            meta=meta,
            marker=dict(symbol="triangle-up", size=10, color=color),
            hovertemplate='x: %{meta[0]} m<br>Force: %{meta[2]} kN/m',
            hoverinfo="skip",
            opacity=0
        )

        trace_2 = go.Scatter(
            x=[x1], y=[0],
            showlegend=False, mode="markers",
            name=name,
            meta=meta,
            marker=dict(symbol="triangle-up", size=10, color=color),
            hovertemplate='x: %{meta[1]} m<br>Force: %{meta[3]} kN/m',
            hoverinfo="skip",
            opacity=0
        )

        if row and col:
            fig.add_trace(trace_1, row=row, col=col)
            fig.add_trace(trace_2, row=row, col=col)
        else:
            fig.add_trace(trace_1)
            fig.add_trace(trace_2)

    return fig


def draw_reaction_hoverlabel(fig, reactions, x_sup, row=None, col=None):
    """Draw a reaction hoverlabel on a plotly figure

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append force representation to.
    load : instance of a load class
        force to be represented on figure
    x_sup : int
        The position for the hoverlabel to be anchored to i.e. the position
        of a support or spring.
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.

    Returns
    -------
    plotly figure
        plotly figure to append hoverlabel representation to.
    """
    # reactions should be [x,y,m]
    x_, y_, m_ = reactions

    # Write hovertemplate depending on support restraints
    hovertemplate = "Reactions"
    if x_:
        hovertemplate += "<br>x: %{meta[0]} kN"
    if y_:
        hovertemplate += "<br>y: %{meta[1]} kN"
    if m_:
        hovertemplate += "<br>m: %{meta[2]} kN.m"
    
    # Create scatter object with opacity 0 for hovertemplate
    trace = go.Scatter(
        x=[x_sup], y=[0],
        showlegend=False, mode="markers",
        name="Reaction",
        meta=reactions,
        marker=dict(symbol="triangle-up", size=10, color='red'),
        hovertemplate=hovertemplate,
        hoverinfo="skip",
        opacity=0
    )
    
    # Add to plot or subplot
    if row and col:
        fig.add_trace(trace, row=row, col=col)
    else:
        fig.add_trace(trace)

    return fig


def draw_support_hoverlabel(fig, support, kx=0, ky=0, row=None, col=None):
    """Draw a reaction hoverlabel on a plotly figure

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append force representation to.
    support : Support instance
        support to be represented on figure
    kx : int, optional
        The spring stiffness of the support in the x direction, default 0
    ky : int, optional
        The spring stiffness of the support in the y direction, default 0
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.


    Returns
    -------
    plotly figure
        plotly figure with hoverlabel appended to it.
    """
    # This could be implemented better (hovertemplates in general could be)
    if kx == oo:
        kx = 0
    if ky == oo:
        ky = 0

    # if ky of kx is > 0 then we are defining a spring label
    # otherwise we are definind a support label

    fixed = support._fixed
    x_sup = support._position

    # Spring
    if kx or ky:
        name = "Spring"
        color = 'orange'
        meta = [kx, ky]
        hovertemplate = "x: %{x} m"
        if kx:
            hovertemplate += "<br>kx: %{meta[0]} kN/mm"
        if ky:
            hovertemplate += "<br>ky: %{meta[1]} kN/mm"
    
    # Support
    else:
        name = "Support"
        color = 'blue'
        meta = [str(fixed)]
        hovertemplate = "x: %{x} m<br>Fixed: %{meta[0]}"

    # necessary for hover information, opacicity 0 so not visible otherwise
    # symbol is arbritrary since invisible
    trace = go.Scatter(
        x=[x_sup], y=[0],
        showlegend=False, mode="markers",
        name=name,
        meta=meta,
        marker=dict(symbol="triangle-up", size=10, color=color),
        hovertemplate=hovertemplate,
        hoverinfo="skip",
        opacity=0
    )

    # Append to plot or subplot.
    if row and col:
        fig.add_trace(trace, row=row, col=col)
    else:
        fig.add_trace(trace)

    return fig


def draw_support_rollers(fig, x_sup, orientation='up', offset=1, row=None,
                         col=None):
    """Draw an anchored group of 3 circles on a plotly figure to represent a
    roller shape.

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append roller shape to.
    x_sup : int
        The x position for the roller shape to be anchored to.
    orientation : 'up' or 'right, optional
        direction that the arrow faces, by default "up"
    offset : int
        offset of the rollers from support location, typically 1 or 0.6,
        default 1.
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.


    Returns
    -------
    plotly figure
        Returns the plotly figure passed into function with the roller shape
        appended to it.
    """
    radius = 1
    if orientation in ['up', 'right']:

        # shifting the position from x_sup to make more aesthetic for 
        # triangle or rectangle.
        # A triangle is wider so an offset of 1 is used.
        # A rectangle uses an offset shorter, currently 0.6 is used.
        shift = offset * -13

        # set centre of circles to be drawn
        if orientation == 'up':
            centres = [(0, shift), (-4, shift), (4, shift)]
        elif orientation == 'right':
            centres = [(shift, 0), (shift, -4), (shift, 4)]

        for xc, yc in centres:

            # Create circle for each point defined in centres
            shape = dict(
                type="circle",
                xref="x", yref="y",
                x0=xc - radius, y0=yc - radius, x1=xc + radius, y1=yc + radius,
                line_color="blue",
                xsizemode='pixel',
                ysizemode='pixel',
                fillcolor='blue',
                xanchor=x_sup,
                yanchor=0
            )

            # Append circle to plot or subplot.
            if row and col:
                fig.add_shape(shape, row=row, col=col)
            else:
                fig.add_shape(shape)

    return fig


def draw_support_spring(fig, support, orientation="up",color='orange',show_values=True, row=None, col=None):
    """Draw an anchored spring shape on a plotly figure.

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append roller shape to.
    support : Support instance
        support to be represented on figure
    orientation : 'up' or 'right, optional
        direction that the arrow faces, by default "up"
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.

    Returns
    -------
    plotly figure
        Returns the plotly figure passed into function with the spring shape
        appended to it."""

    x_sup = support._position

    # x0 and y0 initialised so that when loop through each point in the coords
    # list will have two points to reference.
    x0, y0 = 0, 0

    # reduction of 0.8 used on coords specified (simple reduction modification)
    reduce = 0.8

    if orientation in ['up', 'right']:
        # coords are points between lines to be created
        # label and stiffness are defined for use as meta data to be added to
        # the hovertemplate
        if orientation == 'right':
            coords = [(5, 0), (7, 5), (12, -5), (14, 0), (19, 0)]
            stiffness = support._stiffness[0]
        else:
            coords = [(0, 5), (-5, 7), (5, 12), (0, 14), (0, 19)]
            stiffness = support._stiffness[1]

        # x1 and y1 are the ends of the line to be created
        for x1, y1 in coords:
            x1, y1 = x1 * reduce, y1 * reduce

            # Create dictionary for line shape object. Note: multiple lines
            # added but reference must always be to the same xanchor
            shape = dict(
                type="line",
                xref="x", yref="y",
                x0=x0, y0=y0, x1=x1, y1=y1,
                line_color=color,
                line_width=2,
                xsizemode='pixel',
                ysizemode='pixel',
                xanchor=x_sup,
                yanchor=0
            )

            # Append line to plot or subplot
            if row and col:
                fig.add_shape(shape, row=row, col=col)
            else:
                fig.add_shape(shape)

            # set end point to be start point for the next line
            x0, y0 = x1, y1

        if show_values:
            y0 = max(y0,7)

            annotation = dict(
                xref="x", yref="y",
                x=x_sup,
                y=0,
                yshift=y0*1.5,
                xshift=x0*2,
                text=stiffness,
                font_color=color,
                showarrow=False,
                )

            # Append shape to plot or subplot
            if row and col:
                fig.add_annotation(annotation, row=row, col=col)
            else:
                fig.add_annotation(annotation)


    return fig


def draw_support(fig, support, row=None, col=None):
    """Draw a support on a plotly figure.

    Parameters
    ----------
    fig : plotly figure
        plotly figure to append roller shape to.
    support : Support instance
        support to be represented on figure
    row : int or None,
        Row of subplot to draw line on. If None specified assumes a full plot,
        by default None.
    col : int or None,
        Column of subplot to draw line on. If None specified assumes a full
        plot, by default None.


    Returns
    -------
    plotly figure
        Returns the plotly figure passed into function with a support drawn."""

    # grab values from support instance
    x_sup = support._position
    fixed = support._fixed
    DOF = support._DOF

    # DOF has 1 for partial restraint and fixed has 1 for full restraint. They
    # are only different if a spring support exists
    springx = DOF[0] - fixed[0]
    springy = DOF[1] - fixed[1]

    # run through all possible cases for supports and draw.
    # Note: runs through the values in fixed not DOF so as to not represent
    # the spring more than once (helps with clarity)
    if fixed == [0, 0, 0]:
        fig = fig
    else:
        fig = draw_support_hoverlabel(fig, support, row=row, col=col)

        if fixed == [0, 0, 1]:
            fig = draw_moment(
                fig,
                moment = 1,
                x_sup=support._position,
                color='blue',
                show_values=False,
                row=row,
                col=col)

        elif fixed == [0, 1, 0]:
            fig = draw_support_rollers(
                fig,
                x_sup=support._position,
                orientation='up',
                offset=1,
                row=row,
                col=col)
            fig = draw_support_triangle(
                fig,
                x_sup=support._position,
                orientation='up',
                row=row,
                col=col)

        elif fixed == [1, 0, 0]:
            fig = draw_support_rollers(
                fig,
                x_sup=support._position,
                orientation='right',
                offset=1,
                row=row,
                col=col)
            fig = draw_support_triangle(
                fig,
                x_sup=support._position,
                orientation='right',
                row=row,
                col=col)

        elif fixed == [0, 1, 1]:
            fig = draw_support_rollers(
                fig,
                x_sup=support._position,
                orientation='up',
                offset=0.6,
                row=row,
                col=col)
            fig = draw_support_rectangle(
                fig,
                x_sup=support._position,
                orientation='up',
                row=row,
                col=col)

        elif fixed == [1, 0, 1]:
            fig = draw_support_rollers(
                fig,
                x_sup=support._position,
                orientation='right',
                offset=0.6,
                row=row,
                col=col)
            fig = draw_support_rectangle(
                fig,
                x_sup=support._position,
                orientation='right',
                row=row,
                col=col)
        elif fixed == [1, 1, 0]:
            fig = draw_support_triangle(
                fig,
                x_sup=support._position,
                orientation='up',
                row=row,
                col=col)
        elif fixed == [1, 1, 1]:
            fig = draw_support_rectangle(
                fig,
                x_sup=support._position,
                orientation='right',
                row=row,
                col=col)

        else:
            raise ValueError(
                f"{fixed} does not match the expected support fixed \
                    for support at {x_sup}")

    # if springx then draw a spring with right orientation (value for
    # orientation would be better written as 'x' but wanted to maintain
    # convention used)
    if springx:
        fig = draw_support_spring(
            fig,
            support,
            orientation="right",
            row=row,
            col=col)

    # if springy then draw a spring with up orientation
    if springy:
        fig = draw_support_spring(
            fig, support, orientation="up", row=row, col=col)

    # draw hover lable for springs
    if springx or springy:
        fig = draw_support_hoverlabel(
            fig,
            support,
            kx=support._stiffness[0],
            ky=support._stiffness[1],
            row=row,
            col=col)

    return fig



from distutils.core import setup
import os
import sys

_here = os.path.abspath(os.path.dirname(__file__))

version = {}
with open(os.path.join(_here, 'indeterminatebeam', 'version.py')) as f:
    exec(f.read(), version)


setup(
  name = 'indeterminatebeam',         # How you named your package folder (MyLib)
  packages = ['indeterminatebeam'],   # Chose the same as "name"
  version = version['__version__'],      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'A solver for 1D indeterminate beams',   # Give a short description about your library
  author = 'Jesse Bonanno',                   # Type in your name
  author_email = 'jessebonanno@gmail.com',      # Type in your E-Mail
  url = 'https://github.com/JesseBonanno/IndeterminateBeam',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/JesseBonanno/IndeterminateBeam/archive/'+version['__version__']+'.tar.gz',    # I explain this later on
  keywords = ['statics', 'indeterminate', 'beam', 'civil','structural', 'shear-force','bending-moment','deflection'],   # Keywords that define your package best
  install_requires=[            
          'matplotlib',
          'numpy',
          'sympy>=1.4',
          'plotly>=4.14.1',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)


"""Main module that contains the main class Beam, and auxiliary classes
Support, PointLoadH, PointLoadV, DistributedLoadH, and DistributedLoadV,
PointLoad, PointTorque and TrapezoidalLoad.

Example
-------
>>> beam = Beam(6)
>>> a = Support()
>>> c = Support(6,(0,1,0))
>>> beam.add_supports(a,c)
>>> beam.add_loads(PointLoadV(-15,3))
>>> beam.analyse()
>>> beam.plot()
"""

from collections import namedtuple
import numpy as np
import os
from sympy import (integrate, lambdify, Piecewise, sympify, symbols, 
                   linsolve, sin, cos, oo)
from sympy.abc import x
from math import radians
from indeterminatebeam.data_validation import (assert_number, assert_positive_number,
                             assert_strictly_positive_number)
from indeterminatebeam.plotly_drawing_aid import (
    draw_line, draw_arrowhead, draw_arrow, draw_support_triangle,
    draw_support_rectangle, draw_moment, draw_force, draw_load_hoverlabel,
    draw_reaction_hoverlabel, draw_support_hoverlabel, draw_support_rollers,
    draw_support_spring, draw_support
    )
from plotly.subplots import make_subplots
import plotly.graph_objects as go


class Support:
    """
    A class to represent a support.

    Attributes:
    -------------
        _position: float
            x coordinate of support on a beam (default 0)
        _stiffness: tuple of 3 floats or infinity
            stiffness K (kN/mm) for movement in x, y and bending, oo
            represents infinity in sympy and means a completely fixed
            conventional support, and 0 means free to move.
        _DOF : tuple of 3 booleans
            Degrees of freedom that are restraint on a beam for
            movement in x, y and bending, 1 represents that a reaction
            force exists and 0 represents free (default (1,1,1))
        _fixed: tuple of 3 booleans
            Degrees of freedom that are completely fixed on a beam for
            movement in x, y and bending, 1 represents fixed and 0
            represents free or spring  (default (1,1,1))
        _id : positive number
            id assigned when support associated with Beam object, to
            help remove supports.

    Examples
    --------
    # Creates a fixed suppot at location 0
    >>> Support(0, (1,1,1))
    # Creates a pinned support at location 5
    >>> Support(5, (1,1,0))
    # Creates a roller support at location 5.54
    >>> Support(5.54, (0,1,0))
    # Creates a y direction spring support at location 7.5
    >>> Support(7.5, (0,1,0), ky = 5)
    """

    def __init__(self, coord=0, fixed=(1, 1, 1), kx=None, ky=None):
        """
        Constructs all the necessary attributes for the Support object.

        Parameters:
        -----------
        coord: float
            x coordinate of support on a beam (default 0)
            (default not 0.0 due to a float precision error that
            previously occured)
        fixed: tuple of 3 booleans
            Degrees of freedom that are fixed on a beam for movement in
            x, y and bending, 1 represents fixed and 0 represents free
            (default (1,1,1))
        kx :
            stiffness of x support (kN/mm), if set will overide the
            value placed in the fixed tuple. (default = None)
        ky : (positive number)
            stiffness of y support (kN/mm), if set will overide the
            value placed in the fixed tuple. (default = None)
        """
        # input validation
        assert_positive_number(coord, 'coordinate')

        if kx:
            assert_positive_number(kx, 'kx')
        if ky:
            assert_positive_number(ky, 'ky')

        for a in fixed:
            if a not in [0, 1]:
                raise ValueError(
                    "The provided fixed parameter, must be a tuple of \
                    booleans of length 3"
                )
        if len(fixed) != 3:
            raise ValueError(
                "The provided fixed parameter, must be a tuple of \
                booleans of length 3"
            )

        # Spring representation, set rigid to infinity instead of 1
        self._stiffness = [oo if a else 0 for a in fixed]

        # If kx or ky has been included override oo value
        if kx:
            self._stiffness[0] = kx
        if ky:
            self._stiffness[1] = ky

        # Assign properties for support
        self._DOF = [int(bool(e)) for e in self._stiffness]
        self._fixed = [
            int(bool(e)) if e == oo else 0 for e in self._stiffness
        ]
        self._position = coord
        self._id = None

    def __str__(self):
        return f"""--------------------------------
        id = {self._id}
        position = {float(self._position)}
        Stiffness_x = {self._stiffness[0]}
        Stiffness_y = {self._stiffness[1]}
        Stiffness_M = {self._stiffness[2]} """

    def __repr__(self):
        if self._id:
            return f"<support, id = {self._id}>"
        return "<Support>"


class PointLoad(namedtuple("PointLoad", "force, coord, angle")):
    """Point load described by a tuple of floats: (force, coord, angle).

    Parameters:
    -----------
    Force: float
        Force in kN
    coord: float
        x coordinate of load on beam
    angle: float (between 0 and 180)
        angle of point load in range 0 to 180 where:
        - 0 degrees is purely horizontal +ve
        - 90 degrees is purely vertical +ve
        - 180 degrees is purely horizontal -ve of force sign specified.


    Examples
    --------
    # 10 kN towards the right at x=9 m
    >>> external_force = PointLoad(10, 9, 90)
    # 30 kN downwards at x=3 m
    >>> external_force = PointLoad(-30, 3, 0)
    >>> external_force
    PointLoad(force=-30, coord=3, angle=0)
    """


class PointLoadV(namedtuple("PointLoadV", "force, coord")):
    """Vertical point load described by a tuple of floats:
    (force, coord).

    Parameters:
    -----------
    force: float
        Force in kN
    coord: float
        x coordinate of load on beam

    Examples
    --------
    # 30 kN downwards at x=3 m
    >>> external_force = PointLoadV(-30, 3)
    >>> external_force
    PointLoadV(force=-30, coord=3)
    """


class PointLoadH(namedtuple("PointLoadH", "force, coord")):
    """Horizontal point load described by a tuple of floats:
    (force, coord).

    Parameters:
    -----------
    Force: float
        Force in kN
    coord: float
        x coordinate of load on beam


    Examples
    --------
    # 10 kN towards the right at x=9 m
    >>> external_force = PointLoadH(10, 9)
    >>> external_force
    PointLoadH(force=10, coord=9)
    """


class DistributedLoadV(namedtuple("DistributedLoadV", "expr, span")):
    """Distributed vertical load, described by its functional form
    and application interval.

    Parameters:
    -----------
    expr: sympy expression
        Sympy expression of the distributed load function expressed
        using variable x which represents the beam x-coordinate.
        Requires quotation marks around expression.
    span: tuple of floats
        A tuple containing the starting and ending coordinate that
         the function is applied to.


    Examples
    --------
    # Linearly growing load for 0<x<2 m
    >>> snow_load = DistributedLoadV("10*x+5", (0, 2))
    # Linearly growing load starting at 5kN/m at 1m coordinate and
    # ending at 15kn/m at 2m coordinate
    >>> trapezoidal_load = DistributedLoadV("-5 + 10 * x", (1,2))
    # Uniformly distributed load of 10 kn/m from 1m to 3m coordinate
    >>> UDL = DistributedLoadV(10, (1,3))
    """


class DistributedLoadH(namedtuple("DistributedLoadH", "expr, span")):
    """Distributed horizontal load, described by its functional form
    and application interval.

    Parameters:
    -----------
    expr: sympy expression
        Sympy expression of the distributed load function expressed
        using variable x which represents the beam x-coordinate.
        Requires quotation marks around expression.
    span: tuple of floats
        A tuple containing the starting and ending coordinate that
        the function is applied to.

        Examples
    --------
    # Linearly growing load for 0<x<2 m
    >>> weight = DistributedLoadH("10*x+5", (0, 2))
    """


class PointTorque(namedtuple("PointTorque", "torque, coord")):
    """Point clockwise torque, described by a tuple of floats:
    (torque, coord).

    Parameters:
    -----------
    torque: float
        Torque in kN.m
    coord: float
        x coordinate of torque on beam

    Examples
    --------
    # 30 kN·m (clockwise) torque at x=4 m
    >>> motor_torque = PointTorque(30, 4)
    """


def TrapezoidalLoad(force=(0, 0), span=(0, 0)):
    """Wrapper for the DistributedLoadV class, used to express a
    trapezoidal load distribution.

    Parameters
    ------------
        force : tuple of two floats
            Describes the starting force value and ending force
            value (kN/m)
        span : tuple of two floats
            Describes the starting coordinate value and ending
            coordinate value (m)

    Examples
    --------
    # Linearly growing load starting at 5kN/m ending at 15kn/m
    >>> trapezoidal_load = TrapezoidalLoad((5,15), (1,2))
    # Equivalent expression created using DistributedLoadV
    >>> trapezoidal_load = DistributedLoadV("-5 + 10 * x", (1,2))
    """

    # Data validation
    if len(force) != 2 or len(span) != 2:
        raise TypeError(
            "A tuple of length 2 is required for both the force \
            and span arguments"
        )

    start_load, end_load = force
    start_coordinate, end_coordinate = span

    assert_number(start_load, "force[0] (The starting force)")
    assert_number(end_load, "force[1] (The ending force)")

    assert_positive_number(
        start_coordinate,
        "span[0] (The starting coordinate) "
    )
    assert_strictly_positive_number(end_coordinate, "span[1] (The ending coordinate)")

    if start_coordinate >= end_coordinate:
        raise ValueError(
            "start coordinate should be less than end coordinate"
        )
    
    # Assemble trapzoidal load as a DistributedLoadV load

    # Case 1 - UDL i.e load is constant
    if end_load == start_load:
        return DistributedLoadV(
            start_load,
            (start_coordinate, end_coordinate)
        )

    # Case 2 - Load is not constant and has a non-zero slope
    else:
        a = (end_load - start_load) / (end_coordinate - start_coordinate)
        b = start_load - start_coordinate * a

        return DistributedLoadV(
            f"{a}*x+{b}", (start_coordinate, end_coordinate))


class Beam:
    """
    Represents a one-dimensional beam that can take axial and
    tangential loads.

    Attributes
    --------------
    _x0 :float
        Left end coordinate of beam. This module always takes this
        value as 0.
    _x1 :float
        Right end coordinate of beam. This module always takes this
        to be the same as the beam span.

    _loads: list
        list of load objects associated with the beam
    _distributed_forces_x: list
        list of distributed forces implemented as piecewise functions
    _distributed_forces_y:
        list of distributed forces implemented as piecewise functions

    _normal_forces: sympy piecewise function
        A sympy function representing the internal axial force as a
        function of x.
    _shear_forces: sympy piecewise function
        A sympy function representing the internal shear force as a
        function of x.
    _bending_moments: sympy piecewise function
        A sympy function representing the internal bending moments
        as a function of x.

    _query: list
        A list containing x coordinates that are to have values
        explicitly written on graphs.
    _supports: list
        A list of support objects associated with the beam.
    _reactions: dictionary of lists
        A dictionary with keys for support positions. Each key is
        associated with a list of forces of the form ['x','y','m']

    _E: float
        Young's Modulus of the beam (N/mm2 or MPa)
    _I: float
        Second Moment of Area of the beam (mm4)
    _A: float
        Cross-sectional area of the beam (mm2)

    Notes
    -----
    * The default units for length, force and bending moment (torque)
      are in kN and m (m, kN, kN·m)
    * The default units for beam properties (E, I, A) are in N and mm
      (N/mm2, mm4, mm2)
    * The default unit for spring supports is kN/mm
    * Default properties are for a 150UB18.0 steel beam.
    """

    def __init__(self, span: float = 10, E=2 * 10**5, I=9.05 * 10**6,
                 A=2300):
        """Initializes a Beam object of a given length.

        Parameters
        ----------
        span : float
            Length of the beam span. Must be positive, and the pinned
            and rolling supports can only be placed within this span.
            The default value is 10.
        E: float
            Youngs modulus for the beam. The default value is
            200 000 MPa, which is the youngs modulus for steel.
        I: float
            Second moment of area for the beam about the z axis.
            The default value is 905 000 000 mm4.
        A: float
            Cross-sectional area for the beam about the z axis.
            The default value is 2300 mm4.
        """

        assert_strictly_positive_number(span, 'span')
        assert_strictly_positive_number(E, "Young's Modulus (E)")
        assert_strictly_positive_number(I, 'Second Moment of Area (I)')
        assert_strictly_positive_number(A, 'Area (A)')

        self._x0 = 0
        self._x1 = span

        self._loads = []
        self._distributed_forces_x = []
        self._distributed_forces_y = []

        self._normal_forces = []
        self._shear_forces = []
        self._bending_moments = []

        self._query = []
        self._supports = []
        self._reactions = {}

        self._E = E
        self._I = I
        self._A = A

    def add_loads(self, *loads):
        """Apply an arbitrary list of (point or distributed) loads
        to the beam.

        Parameters
        ----------
        loads : iterable
            An iterable containing DistributedLoad or PointLoad objects
            to be applied to the Beam object. Note that the load
            application point (or segment) must be within the Beam span.

        """
        # For every load validate that has correct values to fit the
        # beam such that no errors occur in analysis from the loads.
        # If load valid then add to self._loads.
        # Note: Have ignored distributedLoadH in this version.
        for load in loads:
            if isinstance(
                load, (
                    PointLoad,
                    PointLoadH,
                    PointLoadV,
                    PointTorque,
                    DistributedLoadV
                )
            ):
                if isinstance(load, DistributedLoadV):
                    left = min(load[1])
                    right = max(load[1])
                    assert_positive_number(left, "span left")
                    assert_positive_number(right, "span right")

                    if self._x0 > left or right > self._x1:
                        raise ValueError(
                            f"Coordinate {load[1]} for {str(load)} is not a point on beam."
                        )

                    elif len(load[1]) != 2:
                        raise ValueError(
                            f"Coordinates for {str(load)} span should only have two numbers entered.")

                    else:
                        self._loads.append(load)

                elif isinstance(
                    load,
                    (PointTorque, PointLoadV, PointLoadH, PointLoad)
                ):
                    assert_number(load[0], 'force')
                    assert_positive_number(load[1], 'x coordinate')

                    if self._x0 > load[1] or load[1] > self._x1:
                        raise ValueError(
                            f"Coordinate {load[1]} for {str(load)} is not a point on beam.")

                    if isinstance(load, PointLoad):
                        assert_number(load[2], 'angle')

                    if abs(round(load[0], 10)) > 0:
                        self._loads.append(load)

        self._update_loads()

    def remove_loads(self, *loads, remove_all=False):
        """Remove an arbitrary list of (point- or distributed) loads
        from the beam.

        Parameters
        ----------
        loads : iterable
            An iterable containing DistributedLoad or PointLoad objects
            to be removed from the Beam object. If object not on beam
            then does nothing.
        remove_all: boolean
            If true all loads associated with beam will be removed.

        """
        # if remove all set, reintialize parameters
        if remove_all:
            self._loads = []
            self._update_loads()
            return None

        # if individual loads check if associated with beam and if so
        # remove. 

        # Could be considered a bug that a user isnt notified
        # if a load isnt removed because it wasnt there. This might be
        # an issue if they dont properly recreate the load they were
        # trying to remove and dont notice that they didnt actually 
        # delete it.

        for load in loads:
            if load in self._loads:
                self._loads.remove(load)

        self._update_loads()

    def add_supports(self, *supports):
        """Apply an arbitrary list of supports (Support objects) to the
        beam.

        Parameters
        ----------
        supports : iterable
            An iterable containing Support objects to be applied to 
            the Beam object. Note that the load application point 
            (or segment) must be within the Beam span.

        """

        # Check support valid then append to self._supports
        for support in supports:
            if not isinstance(support, Support):
                raise TypeError("support must be of type class Support")

            if (self._x0 > support._position) or (
                    support._position > self._x1):
                raise ValueError("Not a point on beam")

            if self._supports == []:
                self._supports.append(support)

            elif support._position not in [
                x._position for x in self._supports
            ]:
                self._supports.append(support)
            else:
                raise ValueError(
                    f"This coordinate {support._position} already has a support associated with it")

    def remove_supports(self, *supports, remove_all=False):
        """Remove an arbitrary list of supports (Support objects) from
        the beam.

        Parameters
        ----------
        *supports : iterable
            An iterable containing Support objects to be removed from
            the Beam object. If support not on beam then does nothing.
        remove_all: boolean
            If true all supports associated with beam will be removed.

        """
        if remove_all:
            self._supports = []
            return None

        for support in self._supports:
            if support in supports:
                self._supports.remove(support)
    
        # Could be considered a bug that a user isnt notified
        # if a support isnt removed because it wasnt there. This might 
        # be an issue if they dont properly recreate the support they 
        # were trying to remove and dont notice that they didnt actually 
        # delete it.

    def get_support_details(self):
        """Print out a readable summary of all supports on the beam. """

        print(f"There are {str(len(self._supports))} supports:", end='\n\n')
        for support in self._supports:
            print(support, end='\n\n')

    def check_determinancy(self):
        """Check the determinancy of the beam.

        Returns
        ---------
        int
            < 0 if the beam is unstable
            0 if the beam is statically determinate
            > 0 if the beam is statically indeterminate

        """

        unknowns = np.array([0, 0, 0])
        equations = 3

        # DOF has a 1 where a reaction force is returned, the sum of 
        # this list returns the number of unknowns associated with a 
        # particular support. If consider all supports then have total
        # number of unknowns to solve for.
        for support in self._supports:
            unknowns = np.array(support._DOF) + unknowns

        # If you dont have any horizontal loads then x equilibrium isnt
        # helpful to solve the beam
        if unknowns[0] == 0:
            equations -= 1
        
        # If you dont have any vertical loads then y equilibrium isnt
        # helpful to solve the beam
        if unknowns[1] == 0:
            equations -= 1

        unknowns = sum(unknowns)

        if unknowns == 0:
            return ValueError("No reaction forces exist")

        if unknowns < equations:
            return ValueError("Structure appears to be unstable")

        else:
            self._determinancy = (unknowns - equations)
            return (unknowns - equations)

    def analyse(self):
        """Solve the beam structure for reaction and internal forces  """

        x0, x1 = self._x0, self._x1

        # create unknown sympy variables
        unknowns_x = {a._position: [
            symbols("x_" + str(a._position)),
            a._stiffness[0]
        ]
            for a in self._supports if a._stiffness[0] != 0}
        unknowns_y = {a._position: [
            symbols("y_" + str(a._position)), a._stiffness[1]]
            for a in self._supports if a._stiffness[1] != 0}
        unknowns_m = {a._position: [symbols(
            "m_" + str(a._position)), a._stiffness[2]]
            for a in self._supports if a._stiffness[2] != 0}

        # grab the set of all the sympy unknowns for y and m and change
        # to a list, do same for x unknowns
        unknowns_ym = [a[0] for a in unknowns_y.values()] \
            + [a[0] for a in unknowns_m.values()]

        unknowns_xx = [a[0] for a in unknowns_x.values()]

        # Assert that there are enough supports. Even though it logically
        # works to have no x support if you have no x loading, it works
        # much better in the program and makes the code alot shorter to
        # just enforce that an x support is there, even when there is no
        # load.
        if len(unknowns_xx) < 1:
            raise ValueError(
                'You need at least one x restraint, even if there are \
                no x forces')

        if len(unknowns_ym) < 2:
            raise ValueError(
                'You need at least two y or m restraints, even if there \
                are no y or m forces')

        # locations where x reaction is and order, for indeterminate axial
        # determaintion
        x_0 = [a for a in unknowns_x.keys()]
        x_0.sort()

        # external reaction equations
        F_Rx = sum(
            integrate(load, (x, x0, x1))
            for load in self._distributed_forces_x
        ) \
            + sum(f.force for f in self._point_loads_x()) \
            + sum([a[0] for a in unknowns_x.values()])

        F_Ry = sum(
            integrate(load, (x, x0, x1))
            for load in self._distributed_forces_y
        ) \
            + sum(f.force for f in self._point_loads_y()) \
            + sum([a[0] for a in unknowns_y.values()])

        # moments taken at the left of the beam, anti-clockwise is positive
        M_R = sum(
            integrate(load * x, (x, x0, x1))
            for load in self._distributed_forces_y
        ) \
            + sum(f.force * f.coord for f in self._point_loads_y()) \
            + sum([p * v[0] for p, v in unknowns_y.items()]) \
            + sum(f.torque for f in self._point_torques()) \
            + sum([a[0] for a in unknowns_m.values()])

        # internal beam equations
        C1, C2 = symbols('C1'), symbols('C2')
        unknowns_ym = unknowns_ym + [C1] + [C2]

        # normal forces is same concept as shear forces only no
        # distributed for now.
        N_i = sum(
            self._effort_from_pointload(f)
            for f in self._point_loads_x()
        ) \
            + sum(
                self._effort_from_pointload(PointLoadH(v[0], p))
                for p, v in unknowns_x.items()
        )

        # shear forces, an internal force acting down would be considered
        # positive by adopted convention hence if the sum of forces on
        # the beam are all positive, our internal force would also be
        # positive due to difference in convention
        F_i = sum(
            integrate(load, x)
            for load in self._distributed_forces_y
        ) \
            + sum(
                self._effort_from_pointload(f)
                for f in self._point_loads_y()
        ) \
            + sum(
                self._effort_from_pointload(PointLoadV(v[0], p))
                for p, v in unknowns_y.items()
        )

        # bending moments at internal point means we are now looking left
        # along the beam when we take our moments (vs when we did external
        # external reactions and we looked right). An anti-clockwise moment
        # is adopted as positive internally. Hence we need to consider a
        # postive for our shear forces and negative for our moments by
        # our sign convention
        M_i = sum(
            integrate(load, x, x)
            for load in self._distributed_forces_y
        ) \
            + sum(
                integrate(self._effort_from_pointload(f), x)
                for f in self._point_loads_y()
        ) \
            + sum(
                integrate(self._effort_from_pointload(PointLoadV(v[0], p)), x)
                for p, v in unknowns_y.items()
        ) \
            - sum(
                self._effort_from_pointload(PointTorque(v[0], p))
                for p, v in unknowns_m.items()
        ) \
            - sum(
                self._effort_from_pointload(f)
                for f in self._point_torques()
        )

        # with respect to x, + constants but the constants are the M at fixed
        # supports

        dv_EI = sum(
            integrate(load, x, x, x)
            for load in self._distributed_forces_y
        ) \
            + sum(
            integrate(self._effort_from_pointload(f), x, x)
            for f in self._point_loads_y()
        ) \
            + sum(
            integrate(self._effort_from_pointload(PointLoadV(v[0], p)), x, x)
            for p, v in unknowns_y.items()
        ) \
            - sum(
            integrate(self._effort_from_pointload(PointTorque(v[0], p)), x)
            for p, v in unknowns_m.items()
        ) \
            - sum(
            integrate(self._effort_from_pointload(f), x)
            for f in self._point_torques()
        ) \
            + C1

        v_EI = sum(
            integrate(load, x, x, x, x)
            for load in self._distributed_forces_y
        ) \
            + sum(
            integrate(self._effort_from_pointload(f), x, x, x)
            for f in self._point_loads_y()
        ) \
            + sum(
            integrate(self._effort_from_pointload(PointLoadV(v[0], p)), x, x, x)
            for p, v in unknowns_y.items()
        ) \
            - sum(
            integrate(self._effort_from_pointload(PointTorque(v[0], p)), x, x)
            for p, v in unknowns_m.items()
        ) \
            - sum(
            integrate(self._effort_from_pointload(f), x, x)
            for f in self._point_torques()
        ) \
            + C1 * x + C2

        # equations , create a lsit fo equations
        equations_ym = [F_Ry, M_R]

        # at location that moment is restaint, the slope is known (to be 0,
        # since dont deal for rotational springs)
        for position in unknowns_m.keys():
            equations_ym.append(dv_EI.subs(x, position))

        # at location that y support is restaint the deflection is known (to be
        # F/k)
        for position in unknowns_y.keys():
            equations_ym.append(
                (v_EI.subs(x, position) * 10 ** 12 / (self._E * self._I))
                + (unknowns_y[position][0] / unknowns_y[position][1])
            )

        # equation for normal forces, only for indeterminate in x
        equations_xx = [F_Rx]

        # the extension of the beam will be equal to the spring
        # displacement on right minus spring displacment on left
        if len(x_0) > 1:
            start = x_0[0]
            # dont consider the starting point? only want to look between
            # supports and not at cantilever sections i think
            for position in x_0[1:]:
                equations_xx.append(
                    (
                        sum(
                            integrate(
                                self._effort_from_pointload(f),
                                (x, x_0[0], position)
                            )
                            for f in self._point_loads_x()
                        )
                        + sum(
                            integrate(
                                self._effort_from_pointload(
                                    PointLoadH(v[0], p)),
                                (x, x_0[0], position))
                            for p, v in unknowns_x.items()
                        )
                    )
                    * 10**3 / (self._E * self._A)
                    + unknowns_x[start][0] / unknowns_x[start][1]
                    # represents elongation displacment on right
                    - unknowns_x[position][0] / unknowns_x[position][1]
                )

        # compute analysis with linsolve
        solutions_ym = list(linsolve(equations_ym, unknowns_ym))[0]
        solutions_xx = list(linsolve(equations_xx, unknowns_xx))[0]

        solutions = [a for a in solutions_ym + solutions_xx]

        solution_dict = dict(zip(unknowns_ym + unknowns_xx, solutions))

        self._reactions = {a._position: [0, 0, 0] for a in self._supports}

        # substitue in value instead of variable in functions
        for var, ans in solution_dict.items():
            v_EI = v_EI.subs(var, ans)  # complete deflection equation
            M_i = M_i.subs(var, ans)  # complete moment equation
            F_i = F_i.subs(var, ans)  # complete shear force equation
            N_i = N_i.subs(var, ans)  # complete normal force equation

            # create self._reactions to allow for plotting of reaction
            # forces if wanted and for use with get_reaction method.
            if var not in [C1, C2]:
                vec, num = str(var).split('_')
                position = float(num)
                if vec == 'x':
                    i = 0
                elif vec == 'y':
                    i = 1
                else:
                    i = 2
                self._reactions[position][i] = float(round(ans, 5))

        # moment unit is kn.m, dv_EI kn.m2, v_EI Kn.m3 --> *10^3, *10^9
        # to get base units. EI unit is N/mm2 , mm4 --> N.mm2
        self._shear_forces = F_i
        self._bending_moments = M_i
        # a positive moment indicates a negative deflection, i thought??
        self._deflection_equation = v_EI * 10 ** 12 / (self._E * self._I)
        self._normal_forces = N_i

    # SECTION - QUERY VALUE
    def get_reaction(self, x_coord, direction=None):
        """Find the reactions of a support at position x.

        Parameters
        ----------
        x_coord: float
            The x_coordinates on the beam to be substituted into the
            equation. List returned (if bools all false)
        direction: str ('x','y' or 'm')
            The direction of the reaction force to be returned.
            If not specified all are returned in a list.

        Returns
        --------
        int
            If direction is 'x', 'y', or 'm' will return an integer
            representing the reaction force of the support in that
            direction at location x_coord.
        list of ints
            If direction = None, will return a list of 3 integers,
            representing the reaction forces of the support ['x','y','m']
            at location x_coord.
        None
            If there is no support at the x coordinate specified.
        """

        if not self._reactions:
            print(
                "You must analyse the structure before calling this function"
            )

        assert_positive_number(x_coord, 'x coordinate')

        if x_coord not in self._reactions.keys():
            return None

        directions = ['x', 'y', 'm']

        if direction:
            if direction not in directions:
                raise ValueError(
                    "direction should be the value 'x', 'y' or 'm'")
            else:
                return self._reactions[x_coord][directions.index(direction)]
        else:
            return self._reactions[x_coord]

    # check if sym_func is the sum of the functions already in
    # plot_analytical

    def _get_query_value(self,x_coord,sym_func,return_max=False,
                         return_min=False,return_absmax=False):
        """Find the value of a function at position x_coord.

        Parameters
        ----------
        x_coord: list
            The x_coordinates on the beam to be substituted into the
            equation. List returned (if bools all false)
        sym_func: sympy function
            The function to be analysed
        return_max: bool
            return max value of function if true
        return_min: bool
            return minx value of function if true
        return_absmax: bool
            return absolute max value of function if true

        Returns
        --------
        int
            Max, min or absmax value of the sympy function depending
            on which parameters are set.
        list of ints
            If x-coordinate(s) are specfied value of sym_func at
            x-coordinate(s).

        Notes
        -----
        * Priority of query parameters is return_max, return_min,
          return_absmax, x_coord (if more than 1 of the parameters are
          specified).

        """

        if isinstance(sym_func, list):
            sym_func = sum(sym_func)
        func = lambdify(x, sym_func, "numpy")

        if 1 not in (return_absmax, return_max, return_min):
            if type(x_coord) == tuple:
                return [round(float(func(x_)), 3) for x_ in x_coord]
            else:
                return round(float(func(x_coord)), 3)

        # numpy array for x positions closely spaced (allow for graphing)
        # i think lambdify is needed to let the function work with numpy
        x_vec = np.linspace(self._x0, self._x1, int(1000))
        y_vec = np.array([func(t) for t in x_vec])
        min_ = float(y_vec.min())
        max_ = float(y_vec.max())

        if return_max:
            return round(max_, 3)
        elif return_min:
            return round(min_, 3)
        elif return_absmax:
            return round(max(abs(min_), max_), 3)

    def get_bending_moment(self,*x_coord, return_max=False,
                           return_min=False, return_absmax=False):
        """Find the bending moment(s) on the beam object.

        Parameters
        ----------
        x_coord: list
            The x_coordinates on the beam to be substituted into the
            equation. List returned (if bools all false)
        return_max: bool
            return max value of function if true
        return_min: bool
            return minx value of function if true
        return_absmax: bool
            return absolute max value of function if true

        Returns
        --------
        int
            Max, min or absmax value of the bending moment function
            depending on which parameters are set.
        list of ints
            If x-coordinate(s) are specfied value of the bending moment
            function at x-coordinate(s).

        Notes
        -----
        * Priority of query parameters is return_max, return_min,
          return_absmax, x_coord (if more than 1 of the parameters are
          specified).

        """

        return self._get_query_value(
            x_coord,
            sym_func=self._bending_moments,
            return_max=return_max,
            return_min=return_min,
            return_absmax=return_absmax
        )

    def get_shear_force(self, *x_coord, return_max=False, return_min=False,
                        return_absmax=False):
        """Find the shear force(s) on the beam object.

        Parameters
        ----------
        x_coord: list
            The x_coordinates on the beam to be substituted into the
            equation. List returned (if bools all false)
        return_max: bool
            return max value of function if true
        return_min: bool
            return minx value of function if true
        return_absmax: bool
            return absolute max value of function if true

        Returns
        --------
        int
            Max, min or absmax value of the shear force function
            depending on which parameters are set.
        list of ints
            If x-coordinate(s) are specfied value of the shear force
            function at x-coordinate(s).

        Notes
        -----
        * Priority of query parameters is return_max, return_min,
          return_absmax, x_coord (if more than 1 of the parameters are
          specified).

        """

        return self._get_query_value(
            x_coord,
            sym_func=self._shear_forces,
            return_max=return_max,
            return_min=return_min,
            return_absmax=return_absmax
        )

    def get_normal_force(self, *x_coord, return_max=False, return_min=False,
                         return_absmax=False):
        """Find the normal force(s) on the beam object.

        Parameters
        ----------
        x_coord: list
            The x_coordinates on the beam to be substituted into the
            equation. List returned (if bools all false)
        return_max: bool
            return max value of function if true
        return_min: bool
            return minx value of function if true
        return_absmax: bool
            return absolute max value of function if true

        Returns
        --------
        int
            Max, min or absmax value of the normal force function
            depending on which parameters are set.
        list of ints
            If x-coordinate(s) are specfied value of the normal force
            function at x-coordinate(s).

        Notes
        -----
        * Priority of query parameters is return_max, return_min,
          return_absmax, x_coord (if more than 1 of the parameters are
          specified).

        """
        return self._get_query_value(
            x_coord,
            sym_func=self._normal_forces,
            return_max=return_max,
            return_min=return_min,
            return_absmax=return_absmax
        )

    def get_deflection(self, *x_coord, return_max=False, return_min=False,
                       return_absmax=False):
        """Find the deflection(s) on the beam object.

        Parameters
        ----------
        x_coord: list
            The x_coordinates on the beam to be substituted into the
            equation. List returned (if bools all false)
        return_max: bool
            return max value of function if true
        return_min: bool
            return minx value of function if true
        return_absmax: bool
            return absolute max value of function if true

        Returns
        --------
        int
            Max, min or absmax value of the deflection function depending
            on which parameters are set.
        list of ints
            If x-coordinate(s) are specfied value of the deflection
            function at x-coordinate(s).

        Notes
        -----
        * Priority of query parameters is return_max, return_min,
          return_absmax, x_coord (if more than 1 of the parameters are
          specified).

        """

        return self._get_query_value(
            x_coord,
            sym_func=self._deflection_equation,
            return_max=return_max,
            return_min=return_min,
            return_absmax=return_absmax)

    # SECTION - PLOTTING

    def add_query_points(self, *x_coords):
        """Document the forces on a beam at position x_coord when
        plotting.

        Parameters
        ----------
        x_coord: list
            The x_coordinates on the beam to be queried on plot.

        """
        for x_coord in x_coords:
            if self._x0 <= x_coord <= self._x1:
                self._query.append(x_coord)
            else:
                return ValueError("Not a point on beam")

    def remove_query_points(self, *x_coords, remove_all=False):
        """Remove a query point added by add_query_points function.

        Parameters
        ----------
        x_coord: list
            The x_coordinates on the beam to be removed from query on
            plot.
        remove_all: boolean
            if true all query points will be removed.

        """
        if remove_all:
            self._query = []
            return None

        for x_coord in x_coords:
            if x_coord in self._query:
                self._query.remove(x_coord)
            else:
                return ValueError("Not an existing query point on beam")

    def plot_beam_external(self):
        """A wrapper of several plotting functions that generates a
        single figure with 2 plots corresponding respectively to:

        - a schematic of the loaded beam
        - reaction force diagram

        Returns
        -------
        figure : `plotly.graph_objs._figure.Figure`
            Returns a handle to a figure with the 2 subplots.

        """
        fig = make_subplots(
            rows=2, cols=1,
            shared_xaxes=True,
            vertical_spacing=0.1,
            subplot_titles=("Beam schematic", "Reaction Forces")
        )

        fig = self.plot_beam_diagram(fig=fig, row=1, col=1)
        fig = self.plot_reaction_force(fig=fig, row=2, col=1)

        fig.update_xaxes(title_text='Beam Length (m)', row=2, col=1)

        fig.update_layout(
            height=550,
            width=700,
            title={'text': "Beam External Conditions", 'x': 0.5},
            title_font_size=24,
            showlegend=False,
            hovermode='x')

        return fig

    def plot_beam_internal(
            self,
            reverse_x=False,
            reverse_y=False):
        """A wrapper of several plotting functions that generates a
        single figure with 4 plots corresponding respectively to:

        - normal force diagram,
        - shear force diagram,
        - bending moment diagram, and
        - deflection diagram

        Parameters
        ----------
        reverse_x : bool, optional
            reverse the x axes, by default False
        reverse_y : bool, optional
            reverse the y axes, by default False

        Returns
        -------
        figure : `plotly.graph_objs._figure.Figure`
            Returns a handle to a figure with the 4 subplots.

        """
        fig = make_subplots(
            rows=4,
            cols=1,
            shared_xaxes=True,
            vertical_spacing=0.05,
            subplot_titles=(
                "Normal Force Diagram",
                "Shear Force Diagram",
                "Bending Moment Diagram",
                "Deflection Diagram"
            )
        )

        fig = self.plot_normal_force(
            reverse_x=reverse_x,
            reverse_y=reverse_y,
            fig=fig,
            row=1,
            col=1
        )
        fig = self.plot_shear_force(
            reverse_x=reverse_x,
            reverse_y=reverse_y,
            fig=fig,
            row=2,
            col=1
        )
        fig = self.plot_bending_moment(
            reverse_x=reverse_x,
            reverse_y=reverse_y,
            fig=fig,
            row=3,
            col=1
        )
        fig = self.plot_deflection(
            reverse_x=reverse_x,
            reverse_y=reverse_y,
            fig=fig,
            row=4,
            col=1
        )

        fig.update_xaxes(title_text='Beam Length (m)', row=4, col=1)

        fig.update_layout(
            height=1000,
            width = 700,
            title={'text': "Analysis Results", 'x': 0.5},
            title_font_size=24,
            showlegend=False,
        )

        return fig

    def plot_beam_diagram(self, fig=None, row=None, col=None):
        """Returns a schematic of the beam and all the loads applied on
        it

        Parameters
        ----------
        fig : bool, optional
            Figure to append subplot diagram too. If creating standalone
            figure then None, by default None
        row : int, optional
            row number if subplot, by default None
        col : int, optional
            column number if subplot, by default None

        Returns
        -------
        figure : `plotly.graph_objs._figure.Figure`
            Returns a handle to a figure with the beam schematic.
        """
        # can do point loads as arrows, not sure about directional point
        # loads though, hmm can have a set length for the arrow , say 50,
        # and use trigonometry to find the x and y offset to achieve.
        # for torques idk
        # hoverinfo is skip to not show any default values, hover
        # template is used to show only the x value and to not worry
        # about the y value
        data = go.Scatter(
            x=[self._x0, self._x1],
            y=[0, 0],
            mode='lines',
            name="Beam_",
            line=dict(color='purple', width=2),
            hovertemplate="%{x} m",
            hoverinfo='skip'
        )

        if fig and row and col:
            fig.add_trace(data, row=row, col=col)
            fig.update_yaxes(visible=False, range=[-3, 3], fixedrange=True,row=row,col=col)
        else:
            fig = go.Figure(data=data)
            # Hovermode x makes two hover labels appear if they are at
            # the same point (default setting means only see the last
            # updated point)
            fig.update_layout(
                title_text="Beam Schematic",
                title_font_size=24,
                showlegend=False,
                hovermode='x',
                title_x=0.5)
            fig.update_xaxes(title_text='Beam Length (m)')
            # visible false means y axis doesnt show, fixing range
            # means wont zoom in y direction

            fig.update_yaxes(visible=False, range=[-3, 3], fixedrange=True)

        # for each support append to figure to have the shapes/traces
        # needed for the drawing
        if row and col:
            for support in self._supports:
                fig = draw_support(fig, support, row=row, col=col)

            for load in self._loads:
                fig = draw_force(fig, load, row=row, col=col)
                fig = draw_load_hoverlabel(fig, load, row=row, col=col)
        else:
            for support in self._supports:
                fig = draw_support(fig, support)

            for load in self._loads:
                fig = draw_force(fig, load)
                fig = draw_load_hoverlabel(fig, load)

        return fig

    def plot_reaction_force(self, fig=None, row=None, col=None):
        """Returns a plot of the beam with reaction forces.

        Parameters
        ----------
        fig : bool, optional
            Figure to append subplot diagram too. If creating standalone
            figure then None, by default None
        row : int, optional
            row number if subplot, by default None
        col : int, optional
            column number if subplot, by default None

        Returns
        -------
        figure : `plotly.graph_objs._figure.Figure`
            Returns a handle to a figure with reaction forces.
        """
        # if a figure is passed it is for the subplot
        # append everything to it rather than creating a new plot.
        data = go.Scatter(
            x=[self._x0, self._x1],
            y=[0, 0],
            mode='lines',
            name="Beam",
            line=dict(color='purple', width=2),
            hovertemplate="%{x} m",
            hoverinfo='skip'
        )

        if fig and row and col:
            fig.add_trace(data, row=row, col=col)
            fig.update_yaxes(visible=False, range=[-3, 3], fixedrange=True,row=row,col=col)

        else:
            fig = go.Figure(data=data)

            # Hovermode x makes two hover labels appear if they are at
            # the same point (default setting means only see the last
            # updated point)
            fig.update_layout(
                title_text="Reaction Forces",
                title_font_size=24,
                showlegend=False,
                hovermode='x',
                title_x=0.5)

            fig.update_xaxes(title_text='Beam Length (m)')

        # visible false means y axis doesnt show, fixing range means
        # wont zoom in y direction
            fig.update_yaxes(visible=False, range=[-3, 3], fixedrange=True)

        for position, values in self._reactions.items():
            x_ = round(values[0], 3)
            y_ = round(values[1], 3)
            m_ = round(values[2], 3)

            #if there are reaction forces
            if abs(x_) > 0 or abs(y_)>0 or abs(m_) > 0:
                #subplot case
                if row and col:
                    fig = draw_reaction_hoverlabel(
                        fig,
                        reactions=[x_, y_, m_],
                        x_sup=position,
                        row=row,
                        col=col
                    )

                    if abs(x_) > 0:
                        fig = draw_force(
                            fig,
                            PointLoadH(x_, position),
                            row=row,
                            col=col
                        )
                    if abs(y_) > 0:
                        fig = draw_force(
                            fig,
                            PointLoadV(y_, position),
                            row=row,
                            col=col)
                    if abs(m_) > 0:
                        fig = draw_force(
                            fig,
                            PointTorque(m_, position),
                            row=row,
                            col=col
                        )
                else:
                    fig = draw_reaction_hoverlabel(
                        fig,
                        reactions=[x_, y_, m_],
                        x_sup=position
                    )

                    if abs(x_) > 0:
                        fig = draw_force(fig, PointLoadH(x_, position))
                    if abs(y_) > 0:
                        fig = draw_force(fig, PointLoadV(y_, position))
                    if abs(m_) > 0:
                        fig = draw_force(fig, PointTorque(m_, position))

        return fig

    def plot_normal_force(self, reverse_x=False, reverse_y=False, fig=None,
                          row=None, col=None):
        """Returns a plot of the normal force as a function of the
        x-coordinate.

        Parameters
        ----------
        reverse_x : bool, optional
            reverse the x axes, by default False
        reverse_y : bool, optional
            reverse the y axes, by default False
        fig : bool, optional
            Figure to append subplot diagram too. If creating standalone
            figure then None, by default None
        row : int, optional
            row number if subplot, by default None
        col : int, optional
            column number if subplot, by default None

        Returns
        -------
        figure : `plotly.graph_objs._figure.Figure`
            Returns a handle to a figure with the normal force diagram.
        """

        xlabel = 'Beam Length'
        ylabel = 'Normal Force'
        xunits = 'm'
        yunits = 'kN'
        title = "Normal Force Plot"
        color = "red"

        fig = self.plot_analytical(
            self._normal_forces,
            color,
            title,
            xlabel,
            ylabel,
            xunits,
            yunits,
            reverse_x,
            reverse_y,
            fig=fig,
            row=row,
            col=col
        )
        return fig

    def plot_shear_force(self, reverse_x=False, reverse_y=False, fig=None,
                         row=None, col=None):
        """Returns a plot of the shear force as a function of the
        x-coordinate.

        Parameters
        ----------
        reverse_x : bool, optional
            reverse the x axes, by default False
        reverse_y : bool, optional
            reverse the y axes, by default False
        fig : bool, optional
            Figure to append subplot diagram too. If creating standalone
            figure then None, by default None
        row : int, optional
            row number if subplot, by default None
        col : int, optional
            column number if subplot, by default None

        Returns
        -------
        figure : `plotly.graph_objs._figure.Figure`
            Returns a handle to a figure with the shear force diagram.
        """

        xlabel = 'Beam Length'
        ylabel = 'Shear Force'
        xunits = 'm'
        yunits = 'kN'
        title = "Shear Force Plot"
        color = "aqua"

        fig = self.plot_analytical(
            self._shear_forces,
            color,
            title,
            xlabel,
            ylabel,
            xunits,
            yunits,
            reverse_x,
            reverse_y,
            fig=fig,
            row=row,
            col=col
        )

        return fig

    def plot_bending_moment(self, reverse_x=False, reverse_y=False, fig=None,
                            row=None, col=None):
        """Returns a plot of the bending moment as a function of the
        x-coordinate.

        Parameters
        ----------
        reverse_x : bool, optional
            reverse the x axes, by default False
        reverse_y : bool, optional
            reverse the y axes, by default False
        fig : bool, optional
            Figure to append subplot diagram too. If creating standalone
            figure then None, by default None
        row : int, optional
            row number if subplot, by default None
        col : int, optional
            column number if subplot, by default None

        Returns
        -------
        figure : `plotly.graph_objs._figure.Figure`
            Returns a handle to a figure with the bending moment diagram.
        """

        xlabel = 'Beam Length'
        ylabel = 'Bending Moment'
        xunits = 'm'
        yunits = 'kN.m'
        title = "Bending Moment Plot"
        color = "lightgreen"
        fig = self.plot_analytical(
            self._bending_moments,
            color,
            title,
            xlabel,
            ylabel,
            xunits,
            yunits,
            reverse_x,
            reverse_y,
            fig=fig,
            row=row,
            col=col)

        return fig

    def plot_deflection(self, reverse_x=False, reverse_y=False, fig=None,
                        row=None, col=None):
        """Returns a plot of the beam deflection as a function of the
        x-coordinate.

        Parameters
        ----------
        reverse_x : bool, optional
            reverse the x axes, by default False
        reverse_y : bool, optional
            reverse the y axes, by default False
        fig : bool, optional
            Figure to append subplot diagram too. If creating standalone
            figure then None, by default None
        row : int, optional
            row number if subplot, by default None
        col : int, optional
            column number if subplot, by default None

        Returns
        -------
        figure : `plotly.graph_objs._figure.Figure`
            Returns a handle to a figure with the deflection diagram.
        """

        xlabel = 'Beam Length'
        ylabel = 'Deflection'
        xunits = 'm'
        yunits = 'mm'
        title = "Deflection Plot"
        color = "blue"
        fig = self.plot_analytical(
            self._deflection_equation,
            color,
            title,
            xlabel,
            ylabel,
            xunits,
            yunits,
            reverse_x,
            reverse_y,
            fig=fig,
            row=row,
            col=col
        )

        return fig

    def plot_analytical(self, sym_func, color="blue", title="", xlabel="",
                        ylabel="", xunits="", yunits="", reverse_x=False,
                        reverse_y=False, fig=None, row=None, col=None):
        """
        Auxiliary function for plotting a sympy.Piecewise analytical
        function.

        Parameters
        -----------
        sym_func: sympy function
            symbolic function using the variable x
        color: str
            color to be used for plot, default blue.
        title: str
            title to show above the plot, optional
        xlabel: str
            physical variable displayed on the x-axis. Example: "Length"
        ylabel: str
            physical variable displayed on the y-axis. Example: "Shear
            force"
        xunits: str
            physical unit to be used for the x-axis. Example: "m"
        yunits: str
            phsysical unit to be used for the y-axis. Example: "kN"
        reverse_x : bool, optional
            reverse the x axes, by default False
        reverse_y : bool, optional
            reverse the y axes, by default False
        fig : bool, optional
            Figure to append subplot diagram too. If creating standalone
            figure then None, by default None
        row : int, optional
            row number if subplot, by default None
        col : int, optional
            column number if subplot, by default None


        Returns
        -------
        figure : `plotly.graph_objs._figure.Figure`
            Returns a handle to a figure with the deflection diagram.
        """
        # numpy array for x positions closely spaced (allow for graphing)
        x_vec = np.linspace(self._x0, self._x1, int(1000))
        # transform sympy expressions to lambda functions which can be used to
        # calculate numerical values very fast (with numpy)
        y_lam = lambdify(x, sym_func, "numpy")
        y_vec = np.array([y_lam(t) for t in x_vec])
        # np.array for y values created
        # np.array for y values created
        data = go.Scatter(
            x=x_vec.tolist(),
            y=y_vec.tolist(),
            mode='lines',
            line=dict(color=color, width=1),
            fill='tozeroy',
            name=ylabel,
            hovertemplate="%{x:.3f} <br>%{y:.3f} "
        )

        if row and col and fig:
            fig = fig.add_trace(data, row=row, col=col)
        else:
            fig = go.Figure(data=data)
            fig.update_layout(title_text=title, title_font_size=30)
            fig.update_xaxes(title_text=str(xlabel + " (" + str(xunits) + ")"))

        if row and col:
            fig.update_yaxes(
                title_text=str(ylabel + " (" + str(yunits) + ")"),
                row=row,
                col=col
            )
            fig.update_yaxes(
                autorange="reversed", row=row, col=col
            ) if reverse_y else None
            fig.update_xaxes(
                autorange="reversed", row=row, col=col
            ) if reverse_x else None
        else:
            fig.update_yaxes(title_text=str(ylabel + " (" + str(yunits) + ")"))
            fig.update_yaxes(autorange="reversed") if reverse_y else None
            fig.update_xaxes(autorange="reversed") if reverse_x else None

        for q_val in self._query:
            q_res = self._get_query_value(q_val, sym_func)
            if q_res < 0:
                ay = 40
            else:
                ay = -40

            annotation = dict(
                x=q_val, y=q_res,
                text=f"{str(q_val)}<br>{str(q_res)}",
                showarrow=True,
                arrowhead=1,
                xref='x',
                yref='y',
                ax=0,
                ay=ay
            )
            if row and col:
                fig.add_annotation(annotation, row=row, col=col)
            else:
                fig.add_annotation(annotation)

        return fig

    def _update_loads(self):
        self._distributed_forces_x = [
            self._create_distributed_force(f)
            for f in self._distributed_loads_x()
        ]
        self._distributed_forces_y = [
            self._create_distributed_force(f)
            for f in self._distributed_loads_y()
        ]

    def _create_distributed_force(
            self,
            load: DistributedLoadH or DistributedLoadV,
            shift: bool = True):
        """
        Create a sympy.Piecewise object representing the provided 
        distributed load.

        :param expr: string with a valid sympy expression.
        :param interval: tuple (x0, x1) containing the extremes of the 
        interval on which the load is applied.
        :param shift: when set to False, the x-coordinate in the 
        expression is referred to the left end of the beam, instead 
        of the left end of the provided interval.
        :return: sympy.Piecewise object with the value of the 
        distributed load.
        """
        expr, interval = load
        x0, x1 = interval
        expr = sympify(expr)
        if shift:
            expr.subs(x, x - x0)
        return Piecewise((0, x < x0), (0, x > x1), (expr, True))

    def _effort_from_pointload(
            self, load: PointLoadH or PointLoadV or PointTorque):
        """
        Create a sympy.Piecewise object representing the shear force 
        caused by a point load.

        :param value: float or string with the numerical value of the 
        point load.
        :param coord: x-coordinate on which the point load is applied.
        :return: sympy.Piecewise object with the value of the shear 
        force produced by the provided point load.
        """
        value, coord = load
        return Piecewise((0, x < coord), (value, True))

    def _point_loads_x(self):
        for f in self._loads:
            if isinstance(f, PointLoadH):
                yield f
            elif isinstance(f, PointLoad):
                force, position, angle = f
                # when angle = 0 then force is 1
                force_x = sympify(force * cos(radians(angle))).evalf(10)

                if abs(round(force_x, 3)) > 0:
                    yield PointLoadH(force_x, position)

    def _point_loads_y(self):
        for f in self._loads:
            if isinstance(f, PointLoadV):
                yield f
            elif isinstance(f, PointLoad):
                force, position, angle = f
                # when angle = 90 then force is 1
                force_y = sympify(force * sin(radians(angle))).evalf(10)

                if abs(round(force_y, 3)) > 0:
                    yield PointLoadV(force_y, position)

    def _distributed_loads_x(self):
        for f in self._loads:
            if isinstance(f, DistributedLoadH):
                yield f

    def _distributed_loads_y(self):
        for f in self._loads:
            if isinstance(f, DistributedLoadV):
                yield f

    def _point_torques(self):
        for f in self._loads:
            if isinstance(f, PointTorque):
                yield f

    def __str__(self):
        return f"""--------------------------------
        <Beam>
        length = {self._x0}
        loads = {str(len(self._loads))}"""

    def __repr__(self):
        return f"<Beam({self._x0})>"


if __name__ == "__main__":
    beam = Beam(5)

    a = Support(0, (1, 1, 1))
    b = Support(5, (0, 1, 0))
    beam.add_supports(a, b)

    load_1 = PointLoad(5, 1, 90)
    load_2 = PointTorque(2, 2)
    load_3 = TrapezoidalLoad((0, 1), (0, 1))
    beam.add_loads(load_1, load_2, load_3)

    beam.add_query_points(1, 2, 3)

    beam.analyse()
    fig = beam.plot_reaction_force()
    fig.write_image("./results.pdf")
